"""自动保存Untitled"""

# Will make `.autosave` dir.  TODO 把各个插件的文件读写需求情况写明


import os

if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import time
from idlelib.config import idleConf
from idlelib.runscript import ScriptBinding


class SaveUntitled:
    def __init__(self, parent):
        self.io = parent.io
        self.get_saved = parent.get_saved  # function

        self.sb = ScriptBinding(parent)
        parent.text.bind("<<check-module>>", self.sb.check_module_event)
        parent.text.bind("<<run-module>>", self.run_module_event)

    def run_module_event(self, evt):
        autosave = idleConf.GetOption('main', 'General', 'autosave', type='bool')
        if autosave and not self.io.filename and not self.get_saved():
            self.io.filename = os.path.abspath(time.strftime('.autosave/Untitled@%Y%m%d_%H%M%S.py'))
            os.makedirs('.autosave', exist_ok=True)
            self.io.save(None)
        self.sb.run_module_event(evt)



'''运行选中'''

if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import tkinter as tk
from idlelib.pyshell import PyShell


class RunSelected(tk.Menu):
    def __init__(self, parent):
        if isinstance(parent, PyShell):
            return

        tk.Menu.__init__(self, parent.menubar, tearoff=0)
        self.text = parent.text
        self.flist = parent.flist
        self.rmenu = parent.rmenu
        self.MakeMenu(4)
        self.Binding()
        parent.amenu.insert_cascade(3, label='Run Selected', menu=self) # TODO 单元测试时插入偏差

    def Run(self, mode):
        '''mode: -1, 0, 1'''
        c1 = ['sel.first', 'insert', '1.0'][mode]
        c2 = ['sel.last',  'end', 'insert'][mode]

        code = self.text.get(c1, c2)
        if code.startswith(' '):
            code = 'if 1:\n' + code

        # ref: idlexlib.extensions.RunSelection
        msg = '# Run Region [%s-%s]\n' % (self.text.index(c1), self.text.index(c2))
        shell = self.flist.open_shell()
        console = shell.interp.tkconsole
        console.text.insert('insert', msg)
        shell.interp.runcode(code)
        # TODO 报错位置和真实行号对应

    def MakeMenu(self, pos):
        params = [( 1, 'Run from Cursor'),
                  (-1, 'Run to Cursor'),
                  ( 0, 'Run Selected')]

        self.add_command(label='Run from Cursor', command=lambda: self.Run(1))
        self.add_command(label='Run to Cursor', command=lambda: self.Run(-1))
        self.add_command(label='Run Selected', command=lambda: self.Run(0))

        rmenu = self.rmenu
        rmenu.insert_command(pos, label='Run from Cursor', command=lambda: self.Run(1))
        rmenu.insert_command(pos, label='Run to Cursor', command=lambda: self.Run(-1))
        rmenu.insert_command(pos, label='Run Selected', command=lambda: self.Run(0))
        rmenu.insert_separator(pos)

    def Binding(self):
        text = self.text
        text.bind('<F6>', lambda e: self.Run(-1))
        text.bind('<F7>', lambda e: self.Run(0))
        text.bind('<F8>', lambda e: self.Run(1))




"""带参数运行"""


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


from tkinter.simpledialog import askstring
from idlelib.pyshell import PyShell


class RunArgs:
    def __init__(self, parent):
        if isinstance(parent, PyShell):
            return

        self.text = parent.text
        self.root = parent.root

        parent.add_adv_menu('Run with args', self.Run, sp=True)

    def Run(self): # TODO 有没有更优雅的方法？
        args = askstring('Run with args', 'args:', parent=self.root)
        if args is not None:
            self.text.insert('1.0', "__import__('sys').argv.append({args!r})\n\n".format(args=args))


if __name__ == '__main__': # for test
    import sys
    print(sys.argv)



'''重载文件'''

# TODO F5运行的时候保存了吗？提示文件有更新


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import os
from tkinter.messagebox import askyesno


def mtime(file):
    if file:
        return str(int(os.stat(file).st_mtime * 1e7))  # TODO 查看修改日期我之前是怎么写的？
    else:
        return 0


class AutoReload:
    def __init__(self, parent):
        self.root = parent.root
        self.io = parent.io

        self.mt = mtime(self.io.filename)
        parent.text_frame.bind('<FocusIn>', self.OnFocusIn)
        parent.after_save.append(self.Refresh)

    def Refresh(self):
        self.mt = mtime(self.io.filename)

    def ReloadFile(self): # TODO 重载记忆位置
        self.io.loadfile(self.io.filename)  # TODO rename也可以用这个方法

    def OnFocusIn(self, e):
        mt = mtime(self.io.filename)
        if self.mt != mt:
            self.mt = mt  # 激活窗口最多只提示一次，下一次提示在本地文件再次发生修改
            if askyesno('Refresh?', 'Find text changed, do you need to refresh?', parent=self.root):  # TODO 参考notepad的窗口提示文本
                self.ReloadFile()
            else:
                pass # todo 设置为未保存状态



'''搜索替换工具条'''

# TEST TEXT
# A quick brown fox
# JuMp oVeR THE lazy dog
# 1377
# [a-z]+

# TODO 处理好和纵向滚动条的相对位置问题
# TODO shell中的stdout无法匹配


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import re
import tkinter as tk
from tkinter.messagebox import showinfo


def SelectSpan(text, span, ins):
    c1, c2 = '1.0+%dc' % span[0], '1.0+%dc' % span[1]
    text.tag_remove('sel', '1.0', 'end')
    text.tag_add('sel', c1, c2)
    text.mark_set('insert', [c1, c2][ins])
    text.see(c1)


def PrepFind(pat, repl, isre=False, case=True, word=False):
    flag = re.M
    if not isre:
        pat = re.escape(pat)
        repl = re.escape(repl)
    if not case:
        flag |= re.I
    if word:  # surround "\b" after escape
        pat = r'\b%s\b' % pat
    return pat, repl, flag


class ReplaceBar(tk.Frame):
    def __init__(self, parent):
        tk.Frame.__init__(self, parent.text_frame)

        self.root = parent.root

        self.show = True

        self.text = parent.text

        # TODO 从idle中获取设置
        self.patvar  = tk.StringVar(self, '')     # search pattern
        self.replvar = tk.StringVar(self, '')     # replace string
        self.revar   = tk.BooleanVar(self, False) # regular expression?
        self.casevar = tk.BooleanVar(self, False) # match case?
        self.wordvar = tk.BooleanVar(self, False) # match whole word?
        self.backvar = tk.BooleanVar(self, False) # search backwards?

        self.patvar .trace('w', self.Find) # TODO 绑定active事件
        self.replvar.trace('w', self.Find)
        self.revar  .trace('w', self.Find)
        self.casevar.trace('w', self.Find)
        self.wordvar.trace('w', self.Find)
        self.backvar.trace('w', self.Find)

        t1 = tk.Entry(self, width=8, textvariable=self.patvar)
        t2 = tk.Entry(self, width=8, textvariable=self.replvar)
        tk.Label(self, text='Find:').pack(side='left')
        t1.pack(side='left', fill='x', expand=True)
        tk.Label(self, text='Repl:').pack(side='left')
        t2.pack(side='left', fill='x', expand=True)

        self.tip = tk.Label(self, text=' Match: 0')
        self.tip.pack(side='left')

        tk.Checkbutton(self, text='Cc', variable=self.casevar).pack(side='left')
        tk.Checkbutton(self, text='Wd', variable=self.wordvar).pack(side='left')
        tk.Checkbutton(self, text='Re', variable=self.revar)  .pack(side='left')

        tk.Button(self, relief='groove', text='<<', command=lambda: self.View(0)).pack(side='left')
        tk.Button(self, relief='groove', text='>>', command=lambda: self.View(1)).pack(side='left')
        tk.Button(self, relief='groove', text='Replace', command=self.Replace).pack(side='left')
        tk.Button(self, relief='groove', text='Replace All', command=self.ReplaceSelected).pack(side='left')

        self.text.bind('<<replace-bar-show>>', self.Flip)
        self.text.event_add('<<replace-bar-show>>', '<Key-Escape>') # add event but not clear exist bindings.
        t1.bind('<Escape>', self.Flip)
        t2.bind('<Escape>', self.Flip)

        self.Flip(-1)

    def Flip(self, evt):
        self.show = not self.show
        if self.show:
            self.pack(fill='x', side='bottom')
            self.Find()
        else:
            self.forget()
            self.text.tag_remove('hit', '1.0', 'end')
            self.text.focus()

    def Find(self, *args):
        self.text.tag_remove('hit', '1.0', 'end')
        self.tip.config(text=' Match: 0')

        pat = self.patvar.get()
        if not pat:
            return
        pat, repl, flag = PrepFind(pat, self.replvar.get(), self.revar.get(), self.casevar.get(), self.wordvar.get())

        # s = self.text.get('sel.first', 'sel.last')
        s = self.text.get('1.0', 'end-1c')
        matchs = [m.span() for m in re.finditer(pat, s, flag)]
        self.Highlight(matchs)

        return matchs, pat, repl, flag

    def Highlight(self, matchs):
        for p1, p2 in matchs:
            self.text.tag_add('hit', '1.0+%dc' % p1, '1.0+%dc' % p2)
        self.tip.config(text=' Match: %d' % len(matchs))

    def View(self, next):
        """next: 1 -> forward, 0 -> backward"""
        # TODO 移动光标到选区边缘后继续查找
        self.backvar.set(not next)
        matchs, pat, repl, flag = self.Find()
        if matchs:
            ins = len(self.text.get('1.0', 'insert')) # cursor offset
            now = sorted([p1 for p1, p2 in matchs] + [ins]).index(ins) - 1
            new = (now + next) % len(matchs)
            SelectSpan(self.text, matchs[new], next)
            self.tip.config(text=' Match: %d/%d' % (new + 1, len(matchs)))

    def Replace(self):
        next = not self.backvar.get()
        if self.text.get('sel.first', 'sel.last'):
            self.ReplaceSelected()
        self.View(next)
        # TODO 如果选区未完全匹配表达式也不替换
        # TODO 第一次只匹配不替换

    def ReplaceSelected(self):
        text = self.text

        save = text.index('insert')
        ss = text.get('sel.first', 'sel.last')
        if not ss:
            text.tag_add('sel', '1.0', 'end-1c')

        s1 = text.get('sel.first', 'sel.last')
        matchs, pat, repl, flag = self.Find()
        s2, n = re.subn(pat, repl, s1, flags=flag)

        text.undo_block_start()
        text.delete('sel.first', 'sel.last')
        text.insert('insert', s2) # TODO 光标移动到了最后
        text.undo_block_stop()

        showinfo('Replace', '%d places was replaced in selected text.' % n, parent=self.root)




'''匹配输入和删除成对括号'''

# TODO 选中单侧括号并修改时同时修改匹配括号

if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


LEFT = '([{\'"'
RIGHT = ')]}'
pair = lambda c: ')]}\'"'['([{\'"'.index(c)]


class SmartPairing:
    def __init__(self, parent):
        self.text = parent.text
        for c in LEFT + RIGHT:
            self.text.bind('<%s>' % c, self.PairAdd)  # '<KeyRelease-%s>'%c
        self.text.bind('<BackSpace>', self.PairDelete)

    def PairAdd(self, evt):
        c = evt.char
        text = self.text

        if c in LEFT: # 输入左括号时输入匹配右括号，允许含有选区
            sel = text.tag_ranges('sel')
            left = 'sel.first' if sel else 'insert'
            right = 'sel.last' if sel else 'insert'

            text.undo_block_start()
            text.insert(left, c)
            text.mark_gravity('insert', 'left')
            text.insert(right, pair(c))
            text.mark_gravity('insert', 'right')
            text.undo_block_stop()

            return 'break'

        elif c in RIGHT: # 输入右括号时移动光标但不键入
            if c == text.get('insert'):
                text.mark_set('insert', 'insert+1c')
                return 'break'

    def PairDelete(self, evt): # 删除左括号时同时删除右括号
        text = self.text
        c = text.get('insert-1c')
        if c in LEFT and pair(c) == text.get('insert'):
            text.delete('insert-1c', 'insert+1c')
            return 'break'
